#ifndef map_buffer_h
#define map_buffer_h

#include <vector>
#include "../include/globals.h"
#include "../include/snapshot.h"
#include <iostream>
#include <cstdlib>
#include <fstream>   

using namespace std;

enum Coord_State{
	HAS_VALUE,
	NO_VALUE
};


enum Command{
	CONNECT_POINTS,
	DRAW_POINTS
};

struct Mem_Infos{

private:
	Command command;
	Position p1;
	Position p2;

public:
	Mem_Infos(): p1(), p2(){
	}

	void set_command(Command command){
		this -> command = command;
	}

	Command get_command(){
		return command;
	}

	void set_p1(Position p){
		p1.set_x(p.get_x());
		p1.set_y(p.get_y());
	}

	Position get_p1(){
		return p1;
	}

	void set_p2(Position p){
		p2.set_x(p.get_x());
		p2.set_y(p.get_y());
	}

	Position get_p2(){
		return p2;
	}
};




struct Mem_Overhead{

private:
	Coord_State state;
	Mem_Infos infos;

public:
	Mem_Overhead(): state(NO_VALUE){
	}

	void set_infos(Mem_Infos info){
		infos.set_command(info.get_command());
		infos.set_p1(info.get_p1());
		infos.set_p2(info.get_p2());
	}

	Mem_Infos get_infos(){
		return infos;
	}

	Coord_State get_state(){
		return state;
	}

	void set_state(Coord_State value){
		this -> state = value;
	}

};

struct Infos_Time{

private:
	double time;
	Mem_Infos infos;

public:
	Infos_Time(): time(0), infos(){

	}

	Infos_Time(int time): time(time), infos(){

	}

	Infos_Time(double time, Mem_Infos infos){
		this -> time = time;
		this -> infos = infos;
	}

	void set_infos(Mem_Infos info){
		infos.set_command(info.get_command());
		infos.set_p1(info.get_p1());
		infos.set_p2(info.get_p2());
	}

	Mem_Infos get_infos(){
		return infos;
	}

	double get_time(){
		return time;
	}

	void set_time(double value){
		this -> time = value;
	}

};



class Map_Buffer_Writer{

public:
	Map_Buffer_Writer(){
		this -> mem = SHARED_MEM;
	}

	void enqueue(Mem_Infos* infos){
			Mem_Overhead* next_free_element = 0; 
			if(queue_size == 0){
				next_free_element = (Mem_Overhead*)mem; 
			} else{
				Mem_Overhead* last_element = (Mem_Overhead*)mem + (queue_size -1);
				if((last_element -> get_state())!= HAS_VALUE){
					queue_size = 0;
					next_free_element = (Mem_Overhead*)mem; 
				} else{
					next_free_element = (Mem_Overhead*)mem + queue_size;
				}
			}
			next_free_element -> set_infos(*infos);
			queue_size++;
			next_free_element -> set_state(HAS_VALUE);
	}
private:
	void* mem;
};



class Map_Buffer_Reader{

public:
	Map_Buffer_Reader(): dequeue_size(0){
		this -> mem = SHARED_MEM;
	}

	Mem_Overhead dequeue(){
			Mem_Overhead infos;
			if(dequeue_size >= queue_size){
				dequeue_size = 0;
			} else{
				Mem_Overhead* next_free_element = (Mem_Overhead*)mem + dequeue_size;
				infos = *next_free_element;
				next_free_element -> set_state(NO_VALUE);
				dequeue_size++;
			}
			return infos;
		}

private:
	void* mem;
	int dequeue_size;
};


class Map_Buffer_FileWriter{

public:
	Map_Buffer_FileWriter(): offset(0) {
		inputFile.open("data.bin", ios::out| ios::binary);
	}

	void enqueue(double time, Mem_Infos* infos){
			if(inputFile.is_open()){
				inputFile.seekg(offset, ios_base::beg);
				inputFile.write(reinterpret_cast<const char*>(&time), sizeof (double));
				inputFile.write(reinterpret_cast<const char*>(infos), sizeof (Mem_Infos));
				offset += sizeof (Infos_Time);
			}
		}

	~Map_Buffer_FileWriter(){
		inputFile.close();
	}

private:
	fstream inputFile;
	int offset;
};


class Map_Buffer_FileReader{

public:
	Map_Buffer_FileReader(): offset(0), size(0) {
		inputFile.open("data.bin", ios::in| ios::binary);
		inputFile.seekg(0, ios::end);
		this -> size= (int)inputFile.tellg();
	}

	Infos_Time dequeue(){
			if(inputFile.good()){
				inputFile.seekg(offset, ios_base::beg);
				char info[sizeof(Infos_Time)]; 
				char* single_info = (char*) &info;
				inputFile.read(single_info, sizeof(Infos_Time));
				if(inputFile.tellg() == -1){
					return Infos_Time(-1);
				}
				Infos_Time* ptr = (Infos_Time*) single_info;
				offset = offset + sizeof(Infos_Time);
				return Infos_Time(ptr -> get_time(), ptr -> get_infos());
			}
			return Infos_Time();
		}

	~Map_Buffer_FileReader(){
		inputFile.close();
	}

private:
	fstream inputFile;
	int offset;
	int size;
};


#endif