#ifndef map_manager_h
#define map_manager_h

#define COMPLETE_TURN 	360
#define RIGHT 			1
#define LEFT			-1
#define COORDINATE_AXES 4
#define PI 3.14159265

//#include "../include/map_viewer.h"
#include "../include/snapshot.h"
#include "../include/sensor_controller.h"
#include "../include/globals.h"
#include <iostream>
#include <sys/mman.h>
#include "../include/map_buffer.h"
#include <math.h>  
#include "../include/clock.h"  


using namespace std;
//using namespace sf;

class Map_Manager{

public:

	Map_Manager(int halts): head_direction(LEFT),robot_direction(0), last_measurement(0), file_writer(){

		this -> halts = halts;
		this -> angle = 360/halts;

		this -> points_array = new Position*[halts];
		for(int i = 0; i < halts; i++){
			points_array[i] = new Position();
		}
		

		this ->  map = new bool**[MAP_SIZE];

		for (int i = 0; i < MAP_SIZE; i++){
			bool** array = new bool*[MAP_SIZE];
			for(int j = 0; j < MAP_SIZE; j++){
				array[j] = new bool(0);
			}
			map[i] = array;
		}
		actual_pos = new Position(MAP_SIZE/2,MAP_SIZE/2);
	}

	Snapshot* make_Snapshot(){
		Sensor_Controller_Mock sensor(MOCKED);
		int measure_count = SNAPSHOT_MEASURES/COORDINATE_AXES;

		if(head_direction == RIGHT){
			head_direction = LEFT;
		} else{
			head_direction = RIGHT;
		}

		Snapshot* snapshot = new Snapshot();
		for(int i = 0; i < halts; i++){
			if(i % measure_count == 0){
				int divide = i/ measure_count; 
				sensor.set_Head( head_direction * divide * 90);
			} else{
				sensor.set_Head(head_direction*i*((int)COMPLETE_TURN/halts));
			}
			
			// entfernt die kommastelle und speichert werte
			
			if(head_direction == RIGHT){
				snapshot -> addValue((sensor.measure()/10),(i+robot_direction)%SNAPSHOT_MEASURES);
			} else{
				snapshot -> addValue((sensor.measure()/10), ((-i) + SNAPSHOT_MEASURES) % SNAPSHOT_MEASURES);
			}
			
		}
		
		for(int i = 0; i < halts; i++){
			cout << " " << snapshot -> getValue(i);
		}
		cout << endl;
		if(head_direction == RIGHT){
				sensor.set_Head_to_abs_pos(360, head_direction );
			} else{
				sensor.set_Head_to_abs_pos(-360, head_direction );
			}
		return snapshot;
	}

	


	void evaluate_Snapshot(Snapshot* snapshot){
		int coordinate_achse = halts/4;
		// Wenn i vielfaches dann mache iwas 
		// wenn nicht berechne quadrant
		int direction_snapshot = robot_direction;
		Position* pos = new Position();
		double relative_angle = 0;
		for(int i = 0; i < SNAPSHOT_MEASURES; i++){
			// If we calculate with double and have round errors it will crash here first!!!!!
			relative_angle =  (i % coordinate_achse) * angle;
			if(i % coordinate_achse == 0 || relative_angle >= 90){
				relative_angle = 0;
			}
			if(direction_snapshot < coordinate_achse){

				if(is_coordinate_axe(direction_snapshot)){
					pos -> set_x(actual_pos -> get_x());
					pos -> set_y(actual_pos -> get_y() + snapshot -> getValue(i));
				} else{
					Position p = calculate_point_first_quadrant(snapshot -> getValue(i),relative_angle);
					pos -> set_x(actual_pos -> get_x() - p.get_x());
					pos -> set_y( actual_pos -> get_y() + p.get_y());
				}

			} else if(direction_snapshot < 2*coordinate_achse){

					if(is_coordinate_axe(direction_snapshot)){
						pos -> set_x(actual_pos -> get_x() - snapshot -> getValue(i));
						pos -> set_y(actual_pos -> get_y());
					} else{
						Position p =  calculate_point_second_quadrant(snapshot -> getValue(i),relative_angle);
						pos -> set_x( actual_pos -> get_x() - p.get_x());
						pos -> set_y( actual_pos -> get_y() - p.get_y());
					}

				} else if(direction_snapshot < 3*coordinate_achse){

					if(is_coordinate_axe(direction_snapshot)){
						pos -> set_x(actual_pos -> get_x());
						pos -> set_y(actual_pos -> get_y() -  snapshot -> getValue(i));
					} else{
						Position p = calculate_point_first_quadrant(snapshot -> getValue(i),relative_angle);
						pos -> set_x(actual_pos -> get_x() + p.get_x());
						pos -> set_y(actual_pos -> get_y() - p.get_y());
					}


					} else{

						if(is_coordinate_axe(direction_snapshot)){
							pos -> set_x(actual_pos -> get_x() + snapshot -> getValue(i));
							pos -> set_y(actual_pos -> get_y());
						} else{
							Position p =  calculate_point_second_quadrant(snapshot -> getValue(i),relative_angle);
							pos -> set_x(actual_pos -> get_x() + p.get_x());
							pos -> set_y(actual_pos -> get_y() + p.get_y());
						}

					}
			Position* new_pos = points_array[direction_snapshot];
			new_pos -> set_x(pos -> get_x());
			new_pos -> set_y(pos -> get_y());
			/**
			cout << "Richtung " << direction_snapshot << endl;
			cout << "angle " << relative_angle << endl;
			cout << " Measured " << snapshot -> getValue(i) << endl;
			
			cout << "x " << new_pos -> get_x() << endl;
			cout << "y " << new_pos -> get_y() << endl;
			*/
			direction_snapshot = (direction_snapshot + 1)% halts;
		}
		delete pos;
		//delete snapshot;
		
	}


	Position calculate_point_first_quadrant(int measured_length, double angle){
		// round y up for "better" values at the end
		int y = (int)(ceil(cos(angle * PI/180.0) * measured_length));
		int x = sqrt(measured_length * measured_length - pow(cos(angle * PI/180.0) * measured_length,2));
		return Position(x,y);
	}

	Position calculate_point_second_quadrant(int measured_length, double angle){
		// round x up for "better" values at the end
		int x = (int)(ceil(cos(angle * PI/180.0) * measured_length));
		int y = sqrt(measured_length * measured_length - pow(cos(angle * PI/180.0) * measured_length,2));
		return Position(x,y);
	}

	bool is_coordinate_axe(int direction){
		int coordinate_achse = SNAPSHOT_MEASURES/4;
		if(direction % coordinate_achse == 0){
			return true;
		}
		return false;
	}



	~Map_Manager(){
		for(int i = 0; i < MAP_SIZE; i++){
			//delete [] map[i];
		} 
		//delete [] map;
	}

	void draw_Snapshot_In_Map(){
		int behind = 0;
		if(robot_direction > halts/2){
			behind = robot_direction -halts/2;
		} else{
			behind = robot_direction + halts/2;
		}

		for(int i = 0; i < this -> halts; i++){
			MyClock* clock = MyClock::get_Clock();
			double time = clock -> get_round_time();
			Position* actual_point = points_array[i];
			Position* successor_point = 0;
			if(i != this -> halts -1){
				successor_point = points_array[i+1];
			} else{
				successor_point = points_array[0];
			}
			double vector_length = sqrt(pow(successor_point -> get_x() - actual_point -> get_x(),2) + pow(successor_point -> get_y() - actual_point -> get_y(),2));
			Mem_Infos* info = new Mem_Infos();
			if((i + 1 % halts) != behind && i != behind && vector_length <= ROBOT_DIAGONAL_LENGTH *2){
				info -> set_command(Command::CONNECT_POINTS);
				info -> set_p1(*actual_point);
				info -> set_p2(*successor_point); 
				file_writer.enqueue(time, info);
				bresenham_Algorithm(actual_point -> get_x(), actual_point -> get_y(), successor_point -> get_x(), successor_point -> get_y());
			} else{
				info -> set_command(Command::DRAW_POINTS);
				info -> set_p1(*actual_point);
				info -> set_p2(*successor_point); 
				file_writer.enqueue(time, info);
				*map[actual_point -> get_x()][ actual_point -> get_y()] = true;
			}
			delete info;
		}
	}


/**
	Bresenham algorithm
*/

void bresenham_Algorithm(int x1,
    int y1,
    int  x2,
    int  y2)
{
    int delta_x(x2 - x1);
    // if x1 == x2, then it does not matter what we set here
    signed char const ix((delta_x > 0) - (delta_x < 0));
    delta_x = std::abs(delta_x) << 1;
 
    int delta_y(y2 - y1);
    // if y1 == y2, then it does not matter what we set here
    signed char const iy((delta_y > 0) - (delta_y < 0));
    delta_y = std::abs(delta_y) << 1;

    *map[x1][y1] = true;
    if (delta_x >= delta_y)
    {
        // error may go below zero
        int error(delta_y - (delta_x >> 1));
 
        while (x1 != x2)
        {
            if ((error >= 0) && (error || (ix > 0)))
            {
                error -= delta_x;
                y1 += iy;
            }
            // else do nothing
 
            error += delta_y;
            x1 += ix;

 			*map[x1][y1] = true;
 			
        }
    }
    else
    {
        // error may go below zero
        int error(delta_x - (delta_y >> 1));
 
        while (y1 != y2)
        {
            if ((error >= 0) && (error || (iy > 0)))
            {
                error -= delta_y;
                x1 += ix;
            }
            // else do nothing
 
            error += delta_x;
            y1 += iy;

 			*map[x1][y1] = true;

        }
    }
}

bool actualize_position(){

	Motor_Controller_Mock controller(MOCKED);
	int coordinate_achse = halts/4;
	Sensor_Controller_Mock sensor(MOCKED);
	//sensor.set_Head(robot_direction*((int)COMPLETE_TURN/halts));
	int measured = (int)(sensor.measure()/10);
	
	controller.robot_drive(30,1000); 

	if(measured != MAX_SENSOR_RANGE){

	double actual_angle = angle * (robot_direction % (SNAPSHOT_MEASURES/4));
	//cout <<  " direction " <<  robot_direction << endl;

	if(last_measurement > measured && last_measurement < measured + 10){
		int measure_diff = abs(last_measurement - measured);
		//cout << "last one " << last_measurement <<  "measured " << measured << endl;

		if(robot_direction < coordinate_achse){
					Position p = calculate_point_first_quadrant(measure_diff,actual_angle);
					actual_pos -> set_x(actual_pos -> get_x() - p.get_x());
					actual_pos -> set_y( actual_pos -> get_y() + p.get_y());

			} else if(robot_direction < 2*coordinate_achse){

						Position p =  calculate_point_second_quadrant(measure_diff,actual_angle);
						actual_pos -> set_x( actual_pos -> get_x() - p.get_x());
						actual_pos -> set_y( actual_pos -> get_y() - p.get_y());
					
				} else if(robot_direction < 3*coordinate_achse){

						Position p = calculate_point_first_quadrant(measure_diff,actual_angle);
						actual_pos -> set_x(actual_pos -> get_x() + p.get_x());
						actual_pos -> set_y(actual_pos -> get_y() - p.get_y());
					} else{
							Position p =  calculate_point_second_quadrant(measure_diff,actual_angle);
							actual_pos -> set_x(actual_pos -> get_x() + p.get_x());
							actual_pos -> set_y(actual_pos -> get_y() + p.get_y());
					}
		MyClock* clock = MyClock::get_Clock();
		double time = clock -> get_round_time();
		Mem_Infos* info = new Mem_Infos();
		info -> set_command(Command::DRAW_POINTS);
		info -> set_p1(*actual_pos);
		info -> set_p2(Position()); 
		file_writer.enqueue(time, info);
		delete info;
	}


	if(measured <= ROBOT_DIAGONAL_LENGTH - 10){
		controller.robot_stop();
		return false;
	} else{
		controller.robot_drive(30,1000);
	}
	last_measurement = measured;
	//cout << "actual pos "  <<  actual_pos -> get_x() << " " << actual_pos -> get_y() << endl;

	}
	return true;
	
}



	bool*** getMapPtr(){
		return this -> map;
	}

	Position* getActualPosition(){
		return this -> actual_pos;
	}

	double getAngle(){
		return this -> angle;
	}

	int getRobotDirection(){
		return this -> robot_direction;
	}

	void setRobotDirection(int robot_direction){
		this -> robot_direction = robot_direction;
	}

	void setLastMeasured(int measured){
		this -> last_measurement = measured;
	}

private:

	double angle;
	int halts;
	int head_direction;
	int robot_direction;
	Position** points_array;
	bool*** map;
	Position* actual_pos;
	int last_measurement;
	Map_Buffer_FileWriter file_writer;
};


struct Map_Coords{
	int command;


};



#endif