#ifndef path_algorithm_h
#define path_algorithm_h

#include "../include/globals.h"
#include "../include/map_manager.h"
#include <math.h>

#define COORDINATE_ACHSE_ANGLE 90
#define SECTORS 4
#define AREA_CHECK 4

class Path_Algorithm
{
public:

	Path_Algorithm(Map_Manager* map_manager): map_manager(map_manager), max_distance_value(0){
		max_distance = new int[SNAPSHOT_MEASURES];

		points_new = new Position*[SNAPSHOT_MEASURES];
		for (int i = 0; i < SNAPSHOT_MEASURES; i++){
			Position* p = new Position();
			points_new[i] = p;
		}
	}

	int decide_direction(int startx, int starty){
		double angle = map_manager -> getAngle();
		int sector = (SNAPSHOT_MEASURES / SECTORS); 

		/**
		* calculate Points with max. distance of 255 for the 1. quadrant
		*/
		for(int i = 0; i <= sector; i++){
			Position* p = points_new[i];
			if(i == sector){
				p -> set_x( startx -  MAX_SENSOR_RANGE);
				p -> set_y( starty);
			} else{
				p -> set_x( startx - sin(angle* PI/180.0 * i)*MAX_SENSOR_RANGE);
				p -> set_y( starty + cos(angle* PI/180.0 * i)*MAX_SENSOR_RANGE);
			}
		}

		/**
		* reflect these points at the x-axe
		*/

		int counter = 0;
		for(int i = sector*2; i > sector ; i--){
			Position* p_mirrored = points_new[i];
			Position* p2 = points_new[counter];
			p_mirrored -> set_x(p2 -> get_x());
			p_mirrored -> set_y(starty - (p2 -> get_y() - starty));
			counter++;
		}

		/**
		* reflect all at y-axe
		*/
		counter = 1;
		for(int i = sector* 4 -1; i > sector * 2; i--){
			Position* p_mirrored = points_new[i];
			Position* p2 = points_new[counter];
			p_mirrored -> set_x(startx + ( abs(startx - p2 -> get_x())));
			p_mirrored -> set_y(p2 -> get_y());
			counter++;
		}

		/**
		* calculate distance with bresenham
		*/
		for(int i = 0; i < SNAPSHOT_MEASURES; i++){
			max_distance[i] = bresenham_Algorithm(startx, starty, points_new[i] -> get_x(), points_new[i] -> get_y(), map_manager -> getMapPtr());
		}

		int max_distance_front = 0;
		int max_distance_front_index = 0;

		for(int i = 0; i < SNAPSHOT_MEASURES; i= i+2){
			if((i > sector)&&(i < 3*sector)){
				continue;
			}
			if(max_distance[i] > max_distance_front){
				max_distance_front = max_distance[i];
				max_distance_front_index = i;
			}
		}

		/**
		for(int i = 0; i < SNAPSHOT_MEASURES; i++){
			if((i > sector)&&(i < 3*sector)){
				continue;
			}
			if(max_distance[i] > max_distance_front){
				max_distance_front = max_distance[i];
				max_distance_front_index = i;
			}
		}
		*/
		cout << "max_distance front " << max_distance_front << endl;
		if(max_distance_front<(ROBOT_DIAGONAL_LENGTH*2)){
			int max = 0;
			int max_index = 0;
			for(int i = sector+1; i < 3*sector; i = i+2){
				if(max_distance[i] > max){
					max = max_distance[i];
					max_index = i;
				}
			}
			this -> max_distance_value = max;
			return max_index;
		} else {
			this -> max_distance_value = max_distance_front;
			return max_distance_front_index;
		}
	}


int bresenham_Algorithm(int x1,
    int y1,
    int  x2,
    int  y2,bool*** map)
{
	int startx = x1;
	int starty = y1;
    int delta_x(x2 - x1);
    // if x1 == x2, then it does not matter what we set here
    signed char const ix((delta_x > 0) - (delta_x < 0));
    delta_x = std::abs(delta_x) << 1;
 
    int delta_y(y2 - y1);
    // if y1 == y2, then it does not matter what we set here
    signed char const iy((delta_y > 0) - (delta_y < 0));
    delta_y = std::abs(delta_y) << 1;
    if (delta_x >= delta_y)
    {
        // error may go below zero
        int error(delta_y - (delta_x >> 1));
 
        while (x1 != x2)
        {
            if ((error >= 0) && (error || (ix > 0)))
            {
                error -= delta_x;
                y1 += iy;
            }
            // else do nothing
 
            error += delta_y;
            x1 += ix;
 			if(*map[x1][y1] == true || check_area(x1,y1,map)){
 				return(sqrt(((starty-y1)*(starty-y1))+((startx-x1)*(startx-x1))));
 			}
 			
        }
    }
    else
    {
        // error may go below zero
        int error(delta_x - (delta_y >> 1));
 
        while (y1 != y2)
        {
            if ((error >= 0) && (error || (iy > 0)))
            {
                error -= delta_y;
                x1 += ix;
            }
            // else do nothing
 
            error += delta_x;
            y1 += iy;

            if(*map[x1][y1] == true || check_area(x1,y1,map)){
 				return(sqrt(((starty-y1)*(starty-y1))+((startx-x1)*(startx-x1))));
 			}

        }
    }
    return MAX_SENSOR_RANGE;
}

bool check_area(int x, int y, bool*** map){

	for(int i = x - AREA_CHECK ; i <= x+ AREA_CHECK ; i++){
			for(int j = y - AREA_CHECK; j <= y + AREA_CHECK; j++){
				if(*map[i][j] == true){
					return true;
				}
		}
	}
	return false;
}


	~Path_Algorithm(){
		for(int i = 0; i < SNAPSHOT_MEASURES; i++){
			Position* p = points_new[i];
			delete p;
		}
		delete[] points_new;
		delete max_distance;
	}

	int getMaxDistanceValue(){
		return this -> max_distance_value;
	}

private:
	Position** points_new;
	int* points;
	int* max_distance;
	Map_Manager* map_manager;
	int max_distance_value;
};



#endif