#ifndef sensor_controller_h
#define sensor_controller_h

#include "../lib/ev3dev.h"
#include <thread>
#include <vector>
#include <cstdarg>
#include "../include/globals.h"
#include <iostream>

using namespace ev3dev;
using namespace std;

class Sensor_Controller{

public:
	Sensor_Controller(bool mocked): ultrasonic(INPUT_1),gyro(INPUT_2),sensor_motor(OUTPUT_B){
		if(!mocked){
			ultrasonic.set_mode(ultrasonic_sensor::mode_us_dist_cm);
			gyro.set_mode(gyro_sensor::mode_gyro_ang);
			sensor_motor.set_position(0);
		}
	}

	int measure(){
		if(ultrasonic.connected()){
			return ultrasonic.value();
		} else{
			cout << "No sensor enabled" << endl;
			return 0;
		}
		
		
	}

	int measureAngle(){
		if(gyro.connected()){
			return gyro.value();
		} else{
			cout << "No gyro-sensor enabled" << endl;
			return 0;
		}
		
		
	}

	void resetAngle(){
		if(gyro.connected()){
			gyro.set_mode(gyro_sensor::mode_gyro_rate);
			gyro.set_mode(gyro_sensor::mode_gyro_ang);
		} else{
			cout << "No gyro-sensor enabled" << endl;
		}
		
		
	}

	void set_Head_to_abs_pos(int position, int direction){
		sensor_motor.set_duty_cycle_sp(direction * 15);
		
		sensor_motor.set_position_sp(position);
		int position_reached = 0;

		// Diese Schleife geht sicher, dass auch die korrekte Position erreicht wurde
		// der Context durchlaeuft diese solange bis die Position nicht nur einmal durchlaufen wurden sondern konstant bestehen bleibt, 
		// der Kopft also auf der korrekten Position gehalten hat, Margin ist 1
		while (position_reached != 1000){
			sensor_motor.run_to_abs_pos();
			if(sensor_motor.position() <= position +1 && sensor_motor.position() >= position -1 ){
				position_reached++;
			}
		}
	}


	void set_Head(int position){

		sensor_motor.set_duty_cycle_sp(15);
		
		sensor_motor.set_position_sp(position);
		int position_reached = 0;

		// Diese Schleife geht sicher, dass auch die korrekte Position erreicht wurde
		// der Context durchlaeuft diese solange bis die Position nicht 
		// nur einmal durchlaufen wurden sondern konstant bestehen bleibt, 
		// der Kopft also auf der korrekten Position gehalten hat, Margin ist 1
		while (position_reached != 2000){
			sensor_motor.run_to_abs_pos();
			if(sensor_motor.position() <= position +1 && sensor_motor.position() >= position -1 ){
				position_reached++;
			}
		}
	}

	void set_Head_Standard_Position(int position){
		sensor_motor.set_position(0);
	}

	int getSensorPosition(){
		return sensor_motor.position();
	}

private:
	ultrasonic_sensor ultrasonic;
	gyro_sensor gyro;
	large_motor sensor_motor;

};



class Sensor_Controller_Mock: public Sensor_Controller{

public:


	Sensor_Controller_Mock(bool mocked): Sensor_Controller(mocked), vector_count(0), mocked_position(0){
		mock_sensor = mocked;
		if(mocked){
			// initialise with mocked values
			this -> addMocked_Measurements(&MEASUREMENTS);
		}
	}

	int measure(){

		if(!mock_sensor){
			return Sensor_Controller::measure();
		} 
		else{
			if(mocked_measurements.size() == 0){
				return 0;
			}

			int data = mocked_measurements.at(vector_count);

			if(vector_count != (mocked_measurements.size()) -1){
				vector_count++;
			} else{
				vector_count = 0;
			}
			return data; 	
		}
	}

	void addMocked_Measurements(const vector<int>* measurements){
		mocked_measurements.clear();
		 for( unsigned int i = 0; i < (measurements -> size()); i++){
		 	mocked_measurements.push_back(measurements -> at(i));
		 }
	}

	void set_Head(int position){
		if(!mock_sensor){
			Sensor_Controller::set_Head(position);
		} else{
			mocked_position = position;
		}
	}

	int getSensorPosition(){
		if(!mock_sensor){
			return Sensor_Controller::getSensorPosition();
		}
		return mocked_position;
	}

	void set_Head_to_abs_pos(int position, int direction){
		if(!mock_sensor){
			Sensor_Controller::set_Head_to_abs_pos(position, direction);
		}
	}


	~Sensor_Controller_Mock(){
	}


private:

	std::vector<int> mocked_measurements;
	unsigned int vector_count;
	bool mock_sensor;
	int mocked_position; 

};




#endif